#include <cassert>
#include <iostream>

#include <gp-bnb/bnb.hpp>

static constexpr bool verbose = false;

namespace gp_bnb {

// --------------------------------------------------------------------------------------
// assignment_state implementation.
// --------------------------------------------------------------------------------------

void assignment_state::resize_nodes(size_t count) {
	assigned_partition_.resize(count, illegal_partition);
	//fill_state_.resize(count);
}

void assignment_state::assign(vertex_id node, partition_index p, std::vector<unsigned int> neighbors) {
	assert(assigned_partition_[node] == illegal_partition);
	assert(p >= 0);

	assigned_partition_[node] = p;
	//for each neighbor of node in other partitions: current_objective++
	for(auto neighbor: neighbors){
		if(assigned_partition_[neighbor] == (p xor 1))
			current_objective_++;
	}
	node_count[p]++;
}

void assignment_state::unassign(vertex_id node, std::vector<unsigned int> neighbors) {
	assert(assigned_partition_[node] != illegal_partition);
	auto p = assigned_partition_[node];

	assigned_partition_[node] = illegal_partition;
	//for each neighbor of node in other partitions: current_objective--
	for(vertex_id neighbor : neighbors){
		if(assigned_partition_[neighbor] == (p xor 1))
			current_objective_--;
	}
	node_count[p]--;
}

// --------------------------------------------------------------------------------------
// trail_state implementation.
// --------------------------------------------------------------------------------------

void trail_state::push_node(vertex_id node, partition_index alt) {
	stack_.push_back({node, alt});
}

void trail_state::pop() {
	stack_.pop_back();
}

// --------------------------------------------------------------------------------------
// solver implementation.
// --------------------------------------------------------------------------------------


void solver::resize_nodes(size_t count) {
	//nodes_.resize_nodes(count);
	assignment_.resize_nodes(count);
}

void solver::define_graph(graph& g){
	nodes_ = g;
}

void solver::solve() {
	//anfangs upper bound?
	best_objective_ = nodes_.num_vertices();


	//partition vector initializieren


	while(true) {
		//falls current sol schlechter als obere grenze: 
		//zähle schritte bis zur nächsten alternative und
		if(assignment_.current_objective() >= best_objective_) {
			int i = trail_.length() - 1;
			while(true) {
				if(i < 0)
					return;
				// Only backtrack to assignments that have an alternative.
				if(trail_.alternative_at(i) != illegal_partition)
					break;
				i--;
			}

			auto node = trail_.node_at(i);
			auto alt = trail_.alternative_at(i);
			//backtracke soviele schritte und
			while(trail_.length() > i){
				backtrack();
			}
			//wende dann die entsprechende alternative an
			expand(node, alt);
		}else if(trail_.length() == nodes_.num_vertices()) {
			//falls wir an einem blatt im suchbaum sind
			// Memorize the new solution.
			best_solution_.resize(nodes_.num_vertices());
			for(vertex_id node = 0; node < nodes_.num_vertices(); node++){
				best_solution_[node] = assignment_.assigned_partition_of(node);
			}
			best_objective_ = assignment_.current_objective();
			std::cerr << "Solution improved to k = " << best_objective_ << std::endl;
		}else{
			//sonst expandiere suchbaum weiter
			int node = trail_.length();
			expand(node, next_possible_partition(node, 0));
		}
	}
}

void solver::expand(vertex_id node, partition_index p) {
	// Search for an alternative BEFORE assigning the node.
	// Because the node is not assigned, this calculation is easy.
	auto alt = next_possible_partition(node, p + 1);
	if(verbose) {
		std::cerr << "Assign node " << node << " to partition " << p << std::endl;
		std::cerr << "    Alternative " << alt << std::endl;
	}

	std::vector<unsigned int> neighbors = nodes_.get_adjacency(node);
	
	assignment_.assign(node, p, neighbors);
	trail_.push_node(node, alt);
}

void solver::backtrack() {
	assert(trail_.length());

	auto node = trail_.node_at(trail_.length() - 1);
	if(verbose)
		std::cout << "Unassign node " << node << std::endl;

	trail_.pop();
	assignment_.unassign(node, nodes_.get_adjacency(node));
}

// Finds the next partition in which the given node can be placed.
partition_index solver::next_possible_partition(vertex_id node, partition_index m) {
	// This function is only correct if the node is not assigned yet.
	assert(assignment_.assigned_partition_of(node) == illegal_partition);

	// partition_count_: Index+1 of last partition that we will try.
	partition_index p = m;

	//falls in einer partition schon die hälfte der knoten: nicht als alt wählen
	if((assignment_.node_count_of(p) * 2) < nodes_.num_vertices()){
		return p;
	}
	else{
		return (p xor 1);
	}
}

} // namespace gp_bnb

